<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli <marcelo@gforgegroup.com>
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */


require_once('SystemUtils.php');
require_once('plugins/scmperforce/lib/exceptions/PerforceException.php');

/**
 * Wrapper for the Perforce binary
 */
class PerforceManager {
	private $perforceBin;
	private $adminName;
	private $adminPassword;
	private $port;
	private $depots = null;
	
	/** Used when creating temp config files */
	const CONFIG_PREFIX = 'gf_perforce_';
	
	const PROTECT_WRITE = 'write';
	const PROTECT_READ = 'read';

	public function __construct() {
		$this->perforceBin = SystemUtils::buildFilePath(GFConf::get('scmperforce.binPath'), 'p4');
		$this->adminName = GFConf::get('scmperforce.adminName');
		$this->adminPassword = GFConf::get('scmperforce.adminPassword');
		$this->port = GFConf::get('scmperforce.port');
	}
	
	/**
	 * Adds or updates the user with the information found in config file $filename
	 * @param	string $filename	File name to read data from
	 */
	public function setUser($filename) {
		$this->executeAdmin('user -i -f < '.$filename);
	}
	
	public function setDepot($filename) {
		$this->executeAdmin('depot -i < '.$filename);
	}
	
	public function setGroup($filename) {
		$this->executeAdmin('group -i < '.$filename);
	}

	/**
	 * Get a list of all the available depots
	 */
	public function getDepots() {
		if (!is_null($this->depots)) {
			return $this->depots;
		}
		
		$this->depots = array();
		
		$output = $this->executeAdmin('depots');
		foreach ($output as $line) {
			list($depot_str, $depot_name, $date, $type, $map) = explode(' ', $line);
			// extract the description
			if (preg_match('/[^\']*\'(.*)\'$/', $line, $matches)) {
				$description = $matches[1];
			} else {
				$description = '';
			}
			
			$depot = array(
						'date' => strtotime($date),
						'type' => $type,
						'map' => $map,
						'description' => $description
					);
			$this->depots[$depot_name] = $depot;
		}
		
		return $this->depots;
	}
	
	/**
	 * @return 	bool
	 */
	public function depotExists($depotName) {
		$depots = $this->getDepots();
		foreach ($depots as $currentDepot => $depot) {
			if ($currentDepot == $depotName) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Deletes the given depot
	 */
	public function deleteDepot($depotName) {
		$this->executeAdmin('obliterate -y //'.$depotName.'/');
		$this->executeAdmin('depot -d '.$depotName);
		
		if (array_key_exists($depotName, $this->depots)) {
			unset($this->depots[$depotName]);
		}
	}
	
	public function setProtect($mode, $group_user, $name, $host, $path) {
		$protectData = $this->getProtect();
		
		$n = count($protectData);
		for ($i = 0; $i < $n; $i++) {
			if ($protectData[$i]['mode'] == $mode &&
					$protectData[$i]['group_user'] == $group_user &&
					$protectData[$i]['name'] == $name &&
					$protectData[$i]['host'] == $host &&
					$protectData[$i]['path'] == $path) {
				// Already has the entry, do nothing
				return;
			}
		}
		
		$protectData[] = array(
						'mode' => $mode,
						'group_user' => $group_user,
						'name' => $name,
						'host' => $host,
						'path' => $path
						);
		
		// Now store the new data
		$protections = array();
		
		foreach ($protectData as $data) {
			//$protections[] = implode(' ', $data);
			$protections[] = $data['mode'].' '.$data['group_user'].' '.$data['name'].' '.$data['host'].' '.$data['path'];
		}
		
		$confName = tempnam('/tmp', self::CONFIG_PREFIX);
		$confFile = new PerforceConfigFile($confName);
		$confFile->setVar('Protections', $protections);
		$confFile->save();
		
		$this->executeAdmin('protect -i < '.$confName);
		
		@unlink($confName);
	}
	
	public function getProtect() {
		// Get the protect data
		$lines = $this->executeAdmin('protect -o');
		
		$protectData = array();
		$i = 0;
		while (true) {
			$line = trim($lines[$i]);
			$i++;
			if (strpos($line, 'Protections:') === 0) {
				// found the "Protections:" line
				break;
			}			
		}
		
		$n = count($lines);
		for (;$i < $n; $i++) {
			$line = trim($lines[$i]);
			if (empty($line)) continue;					// empty line
			if (strpos($line, '#') === 0) continue;		// comment
			
			list($mode, $group_user, $name, $host, $path) = explode(' ', $line);
			$protectData[] = array(
							'mode' => trim($mode),
							'group_user' => trim($group_user),
							'name' => trim($name),
							'host' => trim($host),
							'path' => trim($path)
							);
		}
		
		return $protectData;
	}
	
	/**
	 * Retrieves normalized data for a certain changelist
	 */
	public function getChangelistData($changelist_num) {
		$res = array();
		
		$lines = $this->executeAdmin('describe -s '.escapeshellarg($changelist_num));
		// The format of the output looks like this:
		
		// Change $changelist_num by user@workspace on 2006/10/24 11:13:15
		//
		// Log line #1
		// Log line #2
		// ...
		// Log line #n
		//
		// Affected files...
		//
		// ... //depot1/CREDITS#1 edit
		// ... //depot1/LICENSE#2 edit
		// ... //depot2/somefile#2 edit
		// ... //depot3/someotherfile#4 edit
		
		$log = '';
		$i = 2;
		while (true) {
			$line = trim($lines[$i]);
			if (empty($line)) break;
			$log .= $line."\n";;
			$i++;
		}
		$log = trim($log);
		
		$res['log'] = $log;
		
		$res['changes'] = array();
		// Now get the list of affected files, grouped by depot
		$n = count($lines);
		for (; $i < $n; $i++) {
			$line = trim($lines[$i]);
			if (strpos($line, '#') === 0) continue;		// comment
			if (!preg_match('/\\.\\.\\. (.*)#([0-9]+) (.*)$/', $line, $matches)) continue;		// some other line
			
			$fullpath = $matches[1];
			$newrev = $matches[2];
			$oldrev = $newrev-1;
			$action = $matches[3];
			
			// extract the depot and file name from the full path
			if (preg_match('/^\\/\\/([^\/]*)(.*)$/', $fullpath, $matches)) {
				$depot = $matches[1];
				$path = $matches[2];
				
				if (!array_key_exists($depot, $res['changes'])) {
					$res['changes'][$depot] = array();
				}
				
				$res['changes'][$depot][] = array(
												'path' => $path,
												'oldrev' => $oldrev,
												'action' => $action
											);
			}
		}
		
		return $res;
	}
	
	/**
	 * Creates / updates a trigger
	 *
	 * @param	string $name	Name of the trigger
	 * @param	string $type	Type of trigger (action that triggers the script). See the Perforce documentation.
	 * @param	string $path
	 * @param	string $command	Command to execute. The command MUST be quoted.
	 */
	public function setTrigger($name, $type, $path, $command) {
		$triggerData = $this->getTriggers();
		
		// Update/create the entry
		$triggerData[$name] = array(
							'type'		=> $type,
							'path'		=> $path,
							'command'	=> $command
							);

		// Now store the new data
		$triggers = array();

		foreach ($triggerData as $triggerName => $trigger) {
			$triggers[] = $triggerName.' '.$trigger['type'].' '.$trigger['path'].' '.$trigger['command'];
		}

		$confName = tempnam('/tmp', self::CONFIG_PREFIX);
		$confFile = new PerforceConfigFile($confName);
		$confFile->setVar('Triggers', $triggers);
		$confFile->save();
		
		$this->executeAdmin('triggers -i < '.$confName);
		
		@unlink($confName);
	}
	
	/**
	 * Return the list of triggers
	 */
	public function getTriggers() {
		$lines = $this->executeAdmin('triggers -o');
		
		$triggerData = array();
		$i = 0;
		while (true) {
			$line = trim($lines[$i]);
			$i++;
			if (strpos($line, 'Triggers:') === 0) {
				// found the "Triggers:" line
				break;
			}			
		}
		
		$n = count($lines);
		for (;$i < $n; $i++) {
			$line = trim($lines[$i]);
			if (empty($line)) continue;					// empty line
			if (strpos($line, '#') === 0) continue;		// comment
			
			// Format of the line: 3 fields separated by blanks + the command
			if (preg_match('/([^[:blank:]]*)[[:blank:]]+([^[:blank:]]*)[[:blank:]]+([^[:blank:]]*)[[:blank:]]+(.*)$/', $line, $matches)) {
				$trigger_name = $matches[1];
				$triggerData[$trigger_name] = array(
								'type' => $matches[2],
								'path' => $matches[3],
								'command' => $matches[4]
								);
			}
		}
		
		return $triggerData;

	}
	
	/**
	 * Get the diff for all the open files
	 */
	public function getDiff() {
		return $this->executeAdmin('diff -du');
	}
	
	/**
	 * Executes a Perforce command as the admin user
	 */
	private function executeAdmin($command) {
		$cmd = $this->perforceBin.' -p '.$this->port.' -u '.$this->adminName.' -P "'.$this->adminPassword.'" '.$command.' 2>&1';
		
		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		if ($error) {
			throw new PerforceException(implode("\n", $output));
		}
		
		return $output;
	}
}
?>